/* Copyright 2005 The Apache Software Foundation or its licensors, as
 * applicable.
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

#include "httpd.h"
#include "http_config.h"
#include "http_log.h"

#include "apr_strings.h"
#include "apr_network_io.h"

#include "mod_smtpd.h"

module AP_MODULE_DECLARE_DATA smtpd_spamd_module;

typedef struct {
    int enable;

    char *host;
    int   port;

    float reject_level;

} smtpd_spamd_config;

static void *smtpd_spamd_config_new(apr_pool_t *p,
				    server_rec *s)
{
    smtpd_spamd_config *cfg =
	apr_pcalloc(p, sizeof(smtpd_spamd_config));

    cfg->enable = -1;

    cfg->port = -1;
    cfg->host = NULL;

    cfg->reject_level = -1;

    return cfg;
}

static void *smtpd_spamd_config_merge(apr_pool_t *p, void *b,
				      void *n)
{
    smtpd_spamd_config *base = (smtpd_spamd_config *)b;
    smtpd_spamd_config *new = (smtpd_spamd_config *)n;

    smtpd_spamd_config *cfg =
	apr_pcalloc(p, sizeof(smtpd_spamd_config));

    if ( new->enable == -1 ) {
	cfg->enable = base->enable;
    }
    else {
	cfg->enable = new->enable;
    }

    if ( new->port == -1 ) {
	cfg->port = base->port;
    }
    else {
	cfg->port = new->port;
    }

    if ( new->host == NULL ) {
	if ( base->port != NULL ) {
	    cfg->host = apr_pstrdup(p, base->host);
	}
	else {
	    cfg->host = NULL;
	}
    }
    else {
	cfg->host = apr_pstrdup(p, new->host);
    }

    if ( new->reject_level == -1 ) {
	cfg->reject_level = base->reject_level;
    }
    else {
	cfg->reject_level = new->reject_level;
    }

    return cfg;
}

static const char *smtpd_spamd_enable(cmd_parms *cmd,
				      void *din_dconf, int flag)
{
    smtpd_spamd_config *conf =
	ap_get_module_config(cmd->server->module_config, &smtpd_spamd_module);

    conf->enable = flag;

    return NULL;
}

static const char *smtpd_spamd_host(cmd_parms *cmd,
				    void *din_dconf, const char *host)
{
    smtpd_spamd_config *conf =
	ap_get_module_config(cmd->server->module_config, &smtpd_spamd_module);

    conf->host = apr_pstrdup(cmd->pool, host);

    return NULL;
}

static const char *smtpd_spamd_port(cmd_parms *cmd,
				    void *din_dconf, const char *port)
{
    smtpd_spamd_config *conf =
	ap_get_module_config(cmd->server->module_config, &smtpd_spamd_module);

    conf->port = atoi(port);

    return NULL;
}

static const char *smtpd_spamd_reject_level(cmd_parms *cmd,
					    void *din_dconf, const char *reject_level)
{
    smtpd_spamd_config *conf =
	ap_get_module_config(cmd->server->module_config, &smtpd_spamd_module);

    conf->reject_level = strtod(reject_level, (char **)NULL);

    return NULL;
}

static const command_rec smtpd_spamd_cmds[] =
{
    AP_INIT_FLAG("SmtpSpamdEnable", smtpd_spamd_enable, (void *) NULL, RSRC_CONF,
		 "Enable clamav scanning via clamd"),
    AP_INIT_TAKE1("SmtpSpamdHost", smtpd_spamd_host, (void *) NULL, RSRC_CONF,
		  "The host to connect to clamd on"),
    AP_INIT_TAKE1("SmtpSpamdPort", smtpd_spamd_port, (void *) NULL, RSRC_CONF,
		  "The port to connect to clamd on"),
    AP_INIT_TAKE1("SmtpSpamdRejectLevel", smtpd_spamd_reject_level, (void *) NULL, RSRC_CONF,
		  "The amount over the spam the threshold to start bouncing mail"),
    {NULL}
};

smtpd_retcode smtpd_spamd(smtpd_conn_rec *scr, smtpd_return_data *in)
{
    smtpd_spamd_config *conf = ap_get_module_config(scr->s->module_config, &smtpd_spamd_module);
    smtpd_retcode ret = SMTPD_DECLINED;
    apr_pool_t *pool;
    apr_socket_t *sock;
    apr_sockaddr_t *sockaddr;

    char buf[4096];
    char *str;
    apr_size_t buf_len;
    apr_size_t buf_size = sizeof(buf) - 1;

    apr_size_t total_len = 0;

    ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, 0, scr->s,
		 "smtpd_spamd");

    if (conf->enable != 1) {
	return SMTPD_DECLINED;
    }

    apr_pool_create(&pool, scr->p);

    if (apr_sockaddr_info_get(&sockaddr, conf->host, AF_INET, conf->port, 0, pool) != APR_SUCCESS)
    {
	ap_log_error(APLOG_MARK, APLOG_WARNING | APLOG_NOERRNO, 0, scr->s,
		     "[mod_smtpd_spamd]: apr_sockaddr_info_get failed for %s:%d",
		     conf->host, conf->port);
	goto smtpd_spamd_cleanup;
    }

    if (apr_socket_create(&sock, APR_INET, SOCK_STREAM, APR_PROTO_TCP, pool) != APR_SUCCESS)
    {
	ap_log_error(APLOG_MARK, APLOG_WARNING | APLOG_NOERRNO, 0, scr->s,
		     "[mod_smtpd_spamd]: apr_socket_create failed for %s:%d",
		     conf->host, conf->port);
	goto smtpd_spamd_cleanup;
    }

    if (apr_socket_connect(sock, sockaddr) != APR_SUCCESS)
    {
	ap_log_error(APLOG_MARK, APLOG_WARNING | APLOG_NOERRNO, 0, scr->s,
		     "[mod_smtpd_spamd]: apr_socket_connect failed for %s:%d",
		     conf->host, conf->port);
	goto smtpd_spamd_cleanup;
    }

    {
	apr_finfo_t *info = apr_pcalloc(pool, sizeof(apr_finfo_t) );
	apr_file_info_get(info, APR_FINFO_SIZE, scr->transaction->tfp);

	total_len += info->size;
    }

    str = apr_psprintf(pool, "REPORT SPAMC/1.2\r\nContent-length: %ld\r\n\r\n", total_len);
    buf_len = strlen(str);

    apr_socket_send(sock, str, &buf_len);

    smtpd_body_resetpos(scr->transaction, SMTPD_HEADERS);

    while (smtpd_body_getline(scr->transaction, buf, sizeof(buf)) == APR_SUCCESS) {
	buf_len = strlen(buf);
	apr_socket_send(sock, buf, &buf_len);
    }

    apr_socket_shutdown(sock, APR_SHUTDOWN_WRITE);

    memset(buf, 0, sizeof(buf));
    buf_len = buf_size;
    str = buf;

    while (apr_socket_recv(sock, str, &buf_len) == APR_SUCCESS && buf_len > 0 ) {
	str[buf_len] = '\0';
	str += buf_len;
	buf_len = buf_size - (str-buf);
    }

    if (strncmp(buf, "SPAMD/1.1 0 EX_OK", sizeof("SPAMD/1.1 0 EX_OK")-1) != 0) {
	ap_log_error(APLOG_MARK, APLOG_WARNING | APLOG_NOERRNO, 0, scr->s,
		     "[mod_smtpd_spamd]: SPAMD failure: %s", buf);
	goto smtpd_spamd_close_cleanup;
    }

    str = buf;
    str += sizeof("SPAMD/1.1 0 EX_OK");

    while (*str == '\n' || *str == '\r') { str++; }

    if (strncmp(str, "Spam:", sizeof("Spam:")-1) == 0) {
	int spam = 0;
	float score;
	float threshold;

	if (strncmp(str, "Spam: True", sizeof("Spam: True")-1) == 0) {
	    spam = 1;
	}

	if (sscanf(str, "%*[^;]; %f / %f", &score, &threshold) != 2) {
	    ap_log_error(APLOG_MARK, APLOG_WARNING | APLOG_NOERRNO, 0, scr->s,
			 "[mod_smtpd_spamd]: SPAMD parse error on the header: %s", buf);
	    goto smtpd_spamd_close_cleanup;
	}

	if (spam == 1) {

	    if (conf->reject_level >= 0 && score >= (conf->reject_level + threshold)) {
		char *msg = apr_psprintf(in->p, "message was identified as junk mail, score %.2f/%.2f", score, threshold);
		smtpd_respond_oneline(scr, 550, msg);
		ret = SMTPD_DONE;

		ap_log_error(APLOG_MARK, APLOG_WARNING | APLOG_NOERRNO, 0, scr->s,
			     "[mod_smtpd_spamd]: message was identified as junk mail and denying, score %.2f/%.2f", score, threshold);
	    }
	    else {
		ap_log_error(APLOG_MARK, APLOG_WARNING | APLOG_NOERRNO, 0, scr->s,
			     "[mod_smtpd_spamd]: message was identified as junk mail, score %.2f/%.2f", score, threshold);
	    }
	}
	else {
	    ap_log_error(APLOG_MARK, APLOG_WARNING | APLOG_NOERRNO, 0, scr->s,
			 "[mod_smtpd_spamd]: message was not identified as junk mail, score %.2f/%.2f", score, threshold);
	}
    }

 smtpd_spamd_close_cleanup:
    apr_socket_close(sock);

 smtpd_spamd_cleanup:
    apr_pool_destroy(pool);

    return ret;
}

static void register_hooks(apr_pool_t *p)
{
    APR_OPTIONAL_HOOK(smtpd, data_post, smtpd_spamd, NULL, NULL,
		      APR_HOOK_FIRST);
}

module AP_MODULE_DECLARE_DATA smtpd_spamd_module =
{
    STANDARD20_MODULE_STUFF,
    NULL,                          /* dir config creater */
    NULL,                          /* dir merger --- default is to override */
    smtpd_spamd_config_new,   /* server config */
    smtpd_spamd_config_merge, /* merge server config */
    smtpd_spamd_cmds,         /* command apr_table_t */
    register_hooks                 /* register hooks */
};
